/* threadsp v0.9b, functions for preemptive multithreading
 * Copyright 2002, 2003, 2004 Chris Williams <cow3.14@juno.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "threads.h"

/* length of a thread's timeslice in ticks (tick ~= 1/200 sec) */
#define TIMESLICE_LENGTH 4

list_t threads;
thread_t main_thread;
int num_running_threads;
unsigned int int_level;
unsigned char timeslice_ticks;
unsigned int clock;
userintlist_t userints;

void int_init() {}

/* push_all and pop_all are assembly macros */
inline void push_all() {}
inline void pop_all() {}

/* idle is the HALT instruction on the Z80. */
inline void idle() {}

/* turn interrupts on/off */
/* These are used where semaphores would fail to work in the
 * multithreading functions; that is, if semaphores were
 * used in these functions and the thread switcher needs an
 * object which is already locked, it would block and
 * cooperate (effectively calling itself), thereby causing
 * an infinite loop. :-(  */
inline void inton() {}
inline void intoff() {}

inline void threads_init(void)
{
	threads.head = threads.tail = NULL;
	clock = 0;
	
	thread_create(&main_thread, (void *)NULL, sp + sizeof(void *));
	main_thread.state = STATE_RUNNING;
	num_running_threads = 1;
	
	timeslice_ticks = TIMESLICE_LENGTH;
	int_init();
}

void exit(int retval) { sp = main_thread.stack_top - sizeof(void *); }

int thread_create(thread_t *t, void *(*code)(void), void *stack)
{
	if (!t)
		return -1;
	
	t->code = code;
	t->stack_top = stack;
	t->state = STATE_STARTED;
	t->nice = 0;
	t->ticks_accum = 256/2;
	thread_nice(t, 0);
	
	intoff();
	list_append_node(&threads, &t->node);
	inton();
	
	/*thread_yield();*/
	return 0;
}

int thread_nice(thread_t *t, int inc)
{
	int nice;
	
	if (!t)
		t = (thread_t *)threads.head;
	
	intoff();
	nice = t->nice + inc - PRIO_MIN;
	inton();
	
	if (nice < 0)
		nice = PRIO_MIN;
	else if (nice > (PRIO_MAX - PRIO_MIN))
		nice = PRIO_MAX;
	else
		nice += PRIO_MIN;
	
	/* with nice in the range [-20,20], ticks can be
	 * TIMESLICE_LENGTH +/- 2.5
	 * given TIMESLICE_LENGTH = 4 (20ms) for example, ticks can be
	 * [1.5,6.5]
	 */
	intoff();
	t->nice = nice;
	/* more nice = less ticks */
	/* FIXME: these are hard-coded */
	/* the following line MUST be "/ 8", NOT ">> 3"
	 * (so it works for negative values)
	 */
	t->ticks_int = TIMESLICE_LENGTH + -nice / 8;
	t->ticks_frac = -nice * 32;
	inton();
	
	return 0;
}

int thread_cancel(thread_t *t)
{
	if (!t)
		t = (thread_t *)threads.head;
	
	/* exit immediately if the main thread is cancelled */
	if (t == &main_thread)
		exit(0);
	
	intoff();
	if (STATE_RUNNING == t->state)
		--num_running_threads;
	t->state = STATE_CANCELLED;
	
	thread_yield();
	/* if t is the current, thread_yield() will not return here */
	return 0;
}

void thread_suspend(void)
{
	intoff();
	((thread_t *)threads.head)->state = STATE_SUSPENDED;
	--num_running_threads;
	
	thread_yield();
}

unsigned int settimeslice(thread_t *t)
{
	timeslice_ticks = t->ticks_int;
	
	/* has ticks_accum wrapped around? */
	if ((t->ticks_accum += t->ticks_frac) < t->ticks_frac)
		++timeslice_ticks;
	
	return timeslice_ticks;
}

void int_start(void)
{
	userint_t *i;
	((void(*)())0x0066)(); /* default TI-86 int handler */
	
	intoff();
	
	/* call user/other int handlers */
	for (i = userints.head; i; i = i->next)
		i->code();
	
	++clock;
	
	if (!--timeslice_ticks)
		thread_yield();
}

#if 0 /* new scheduler */

/* completely debug this new scheduler before using it */

/* thread_sched
 * returns a pointer to the OLD (current) thread
 * returns 0 if OLD thread == NEW thread (so we don't have to switch).
 * thread.head points to the NEW thread
 */
static thread_t *thread_sched(void)
{
	thread_t *t;
	thread_t *old = (thread_t *)threads.head;
	
	threads.tail = threads.head;
	threads.head = threads.head->next; /* next thread */
	
	while ((t = (thread_t *)threads.head) != old) {
		if (t->state == STATE_RUNNING
		 || t->state == STATE_SUSPENDED
		 || t->state == STATE_STARTED)
			if (settimeslice(t))
				return old;
		
		list_remove_head(&threads);
	}
	
	return (thread_t *)0;
}

static void thread_switch(thread_t *old)
{
	thread_t *t;
	push_all();
	
	t = (thread_t *)threads.head;
	
	old->sp = sp;
	
	if (t->state == STATE_SUSPENDED) {
		if (t == &main_thread && 0 == num_running_threads) {
			/* disable switching for the next int */
			timeslice_ticks = 0;
			inton();
			idle();
			intoff();
		}
		
		t->state = STATE_RUNNING;
		++num_running_threads;
	}
	
	if (t->state == STATE_RUNNING) {
		sp = t->sp;
		pop_all();
		inton();
		return;
	}
	if (t->state == STATE_STARTED) {
		sp = t->stack_top;
		
		t->state = STATE_RUNNING;
		++num_running_threads;
		
		inton();
		thread_exit(t->code());
	}
}

void thread_yield(void)
{
	thread_t *t;
	if (!(t = thread_sched()));
		thread_switch(t);
}

#else /* old scheduler */

static void thread_switch(void)
{
	intoff();
	push_all(); /* save all registers */
	
	((thread_t *)threads.head)->sp = sp;
	
next:	threads.tail = threads.head;
	threads.head = threads.head->next; /* go to next thread in the list */
	
	while (1) {
		thread_t *t = (thread_t *)threads.head;
		switch (t->state) {
		case STATE_RUNNING:
			if (!settimeslice(t))
				goto next;
			
			sp = t->sp; /* go to thread's stack space */
			pop_all(); /* restore all registers */
			inton();
			return; /* return to the running thread */
		case STATE_SUSPENDED:
			if (t == &main_thread && 0 == num_running_threads) {
				/* disable switching for the next int */
				timeslice_ticks = 0;
				inton();
				idle();
				intoff();
			}
			
			t->state = STATE_RUNNING;
			++num_running_threads;
			break;
		case STATE_STARTED:
			if (!settimeslice(t))
				goto next;
			
			sp = t->stack_top; /* go to thread's stack space */
			
			t->state = STATE_RUNNING;
			++num_running_threads;
			
			inton();
			thread_exit(t->code());
		default:
			list_remove_head(&threads);
			/* We don't want to keep a thread who has stopped
			 * (STATE_CANCELLED or STATE_EXITED) or whose state
			 * is invalid. */
		}
	}
}

void thread_yield(void)
{
	thread_switch();
}

#endif /* scheduler */

void thread_exit(void *retval)
{
	thread_t *t = (thread_t *)threads.head;
	
	if (t == &main_thread)
		exit((int)retval);
	
	intoff();
	t->state = STATE_EXITED;
	t->retval = retval;
	--num_running_threads;
	
	thread_yield();
}

void *thread_join(thread_t *t)
{
	void *v;
	
	if (!t)
		return (void *)-1;
	
	while (1) {
		intoff();
		if (STATE_EXITED == t->state) {
			v = t->retval;
			break;
		}
		if (STATE_CANCELLED == t->state) {
			v = (void *)-1;
			break;
		}
		thread_suspend();
	}
	
	inton();
	return v;
}
